/*	
 *	main.cpp
 *
 *	dmalloc - malloc dump utility
 *	Copyright (C) 2002 Naoyuki Sawa
 *
 *	* Tue Jul  2 21:24:00 JST 2002 Naoyuki Sawa
 *	- 쐬JnB
 */
#define STRICT
#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pieceif.h>
#pragma comment(lib,"pieceif.lib")

#define VERSION		"20020703"

/****************************************************************************
 *	萔E^
 ****************************************************************************/

typedef struct tagSYSTEMINFO { /* piece.hQ */
	unsigned short size;
	unsigned short hard_ver;
	unsigned short bios_ver;
	unsigned short bios_date;
	unsigned long sys_clock;
	unsigned short vdde_voltage;
	unsigned short resv1;
	unsigned long sram_top;
	unsigned long sram_end;
	unsigned long pffs_top;
	unsigned long pffs_end;
} SYSTEMINFO;

typedef struct tagHEAPINFO { /* /use/piece/include/stdlib.hQ */
	unsigned long start_alloc; /* ansi_InitMalloc̑ێ  */
	unsigned long end_alloc;   /* ansi_InitMalloc̑ێ  */
	unsigned long nxt_alc_p;   /* 󂫗̈̐擪AhX         */
	unsigned long tbl_ptr;     /* 蓖ăe[u̐擪AhX */
	unsigned long row;          /* 蓖ăe[u̗vf       */
} HEAPINFO;

typedef struct tagBLKTBL { /*  */
	unsigned long ptr;	/* ̃ubN̐擪AhX   */
	unsigned long len:31;	/* ̃ubÑTCY[oCg] */
	unsigned long use: 1;	/* 0:/1:gp              */
} BLKTBL;

/* * ȂubNe[u͈vf]vŁA
 *	tbl_ptr = (BLKTBL*)end_alloc - (row + 1)
 *   ɂȂĂ܂B
 *   ԎႢAhX̃e[uvfɂ̓S~Ă܂B
 * * ׂ荇e[uvf̎wubNAhX͘AĂ܂B
 *	blktbl[n].ptr + blktbl[n].len = blktbl[n - 1].ptr
 *   ɐ܂B
 * * ʃAhX(13FFFF)̃e[uvf̕A
 *   ʃAhX(100000)̃e[uvfɊ蓖Ă̂ŁA
 *   ubNAhX̑ƃe[uvfAhX̑
 *   tł邱ƂɒӂĂB
 */

/****************************************************************************
 *	ϐ
 ****************************************************************************/

SYSTEMINFO sysinfo;	/* VXe       */
char* sram;		/* SRAMS̃C[W   */
int sram_len;		/* SRAMTCY         */

/****************************************************************************
 *	֐錾
 ****************************************************************************/

void usage();
int sram_read();
unsigned long heap_find(unsigned long heapinfo_addr);
void heap_dump(unsigned long heapinfo_addr);

/****************************************************************************
 *	usage()
 ****************************************************************************/

void
usage()
{
	fprintf(stderr, "dmalloc - malloc dump utility (%s)\n", VERSION);
	fprintf(stderr, "Copyright (C) 2002 Naoyuki Sawa <nsawa@north.hokkai.net>\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "USAGE: dmalloc (NO ARGUMENTS, NO OPTIONS!)\n");
	abort();
}

/****************************************************************************
 *	main
 ****************************************************************************/

int
main(int argc, char* argv[])
{
#define N_MAX	1024

	int retval, i, n;
	unsigned long addr, heapinfo_addr[N_MAX];

	if(argc != 1) usage();

	/* SRAMC[Wǂݍ݂܂B */
	retval = sram_read();
	if(retval != 0) abort();

	/* q[vT܂B */
	n = 0;
	addr = sysinfo.sram_top;
	while(n < N_MAX) {
		addr = heap_find(addr);
		if(addr == 0) break;
		heapinfo_addr[n++] = addr;
		addr += 4;
	}
	if(n == 0) {
		fprintf(stderr, "q[v񂪌܂B\n");
		abort();
	}
	if(n != 1) {
		fprintf(stderr, "### xĨq[v񂪌܂B\n");
		/* p */
	}

	/* q[v\܂B */
	for(i = 0; i < n; i++) {
		if(n != 1) printf("\n### %d\n\n", i + 1);
		heap_dump(heapinfo_addr[i]);
	}

	return 0;

#undef N_MAX
}

/****************************************************************************
 *	sram_read
 ****************************************************************************/

static int
ismAppPause(int pausef)
{
	unsigned char tmp[2];
	tmp[0] = 16;
	tmp[1] = (unsigned char)pausef;
	return ismCmdW(tmp, 2, 0, 0);
}

int
sram_read()
{
	int result, retval, ofs, len;

	result = -1;

	/* P/ECEɐڑ܂B */
	retval = ismInitEx(0, 0);
	if(retval != 0) {
		fprintf(stderr, "P/ECEɐڑł܂B\n");
		goto L10;
	}

	/* VXe擾܂B */
	retval = ismGetVersion(&sysinfo, 1);
	if(retval == PIECE_INVALID_VERSION) {
		fprintf(stderr, "VXe񂪎擾ł܂B\n");
		goto L20;
	}

	/* SRAMŜǂݍނ߂̃mۂ܂B */
	sram_len = sysinfo.sram_end - sysinfo.sram_top;
	sram = (char*)malloc(sram_len);
	if(sram == NULL) {
		fprintf(stderr, "słB\n");
		goto L20;
	}

	/* ǂݍݒɓeςȂ悤A|[Y܂B */
	retval = ismAppPause(1);
	if(retval != 0) {
		fprintf(stderr, "AvP[V܂B\n");
		goto L20;
	}

	/* SRAMŜǂݍ݂܂B
	 * x̎MTCY64KB肩瓮sɂȂXA
	 *   P/ECEnOAbv邱Ƃ̂ŁAM邱Ƃɂ܂B
	 *   oN]̈STCY4KBƕƂ̂ŁA4KBɂ܂B
	 */
	ofs = 0;
	while(ofs < sram_len) {
		len = sram_len - ofs;
		if(len > 0x1000) len = 0x1000; /* 4KB */
		retval = ismReadMem(sram + ofs, sysinfo.sram_top + ofs, len);
		if(retval != 0) {
			fprintf(stderr, "SRAM̓ǂݍ݂Ɏs܂B\n");
			goto L30;
		}
		ofs += len;
	}

	result = 0; /* ǂݍݐ */

L30:
	/* |[Y܂B */
	ismAppPause(0);
L20:
	/* P/ECEؒf܂B */
	ismExit();
L10:
	return result;
}

/****************************************************************************
 *	heap_find
 ****************************************************************************/

unsigned long
heap_find(unsigned long heapinfo_addr)
{
	int i;
	unsigned long blktbl_addr;
	unsigned long ptr;
	HEAPINFO hi;
	BLKTBL bt;

	for( ; heapinfo_addr <= sysinfo.sram_end - sizeof(HEAPINFO); heapinfo_addr += 4) {
		memcpy(&hi, &sram[heapinfo_addr - sysinfo.sram_top], sizeof(HEAPINFO));

		/***** HEAPINFOPƂ̌ *****/

		if(hi.start_alloc & 3) continue; /* q[vJnAhX̃ACgG[ */
		if(hi.nxt_alc_p & 3) continue;   /* 󂫗̈擪AhX̃ACgG[ */
		if(hi.tbl_ptr & 3) continue;     /* 蓖ăe[u擪AhX̃ACgG[ */
		if(hi.end_alloc & 3) continue;   /* q[vIAhX̃ACgG[ */

		if(hi.start_alloc < sysinfo.sram_top) continue; /* SRAM擪`q[v擪   */
		if(hi.start_alloc > hi.nxt_alc_p) continue;     /*         `󂫗̈擪 *
		if(hi.nxt_alc_p > hi.tbl_ptr) continue;         /*         `e[u擪 */
		if(hi.tbl_ptr > hi.end_alloc) continue;         /*         `IAhX */
		if(hi.end_alloc > sysinfo.sram_end) continue;   /*         `SRAMI[ ̏ł邱 */

		if(hi.tbl_ptr != hi.end_alloc - sizeof(BLKTBL) * (hi.row + 1)) continue; /* u萔E^vZNṼRgQ */

		/* * HEAPINFOPƂ̌ŁAĂA͂Pɍi܂B
		 *   ̃ubN蓖ĂĂقǁÅm͍Ȃ܂A
		 *   ۂɂ́Aansi_InitMalloc()ŏłAP̐܂B
		 * * Ô߁AāAubNe[ǔsƂɂ܂B
		 */

		/***** BLKTBĽ *****/

		ptr = hi.start_alloc; /* ŏ̃ubN̓q[v擪ɂ͂ */
		for(i = 0; i < (int)hi.row; i++) {
			blktbl_addr = hi.end_alloc - sizeof(BLKTBL) * (i + 1);
			memcpy(&bt, &sram[blktbl_addr - sysinfo.sram_top], sizeof(BLKTBL));
			if(bt.ptr != ptr) break; /* ÕubNƘAĂȂ     */
			if(bt.len & 3) break;    /* 蓖ĒPʂ̃ACgG[ */
			ptr += bt.len;           /* ̃ubNAhX߂Ă */
			if(ptr < hi.start_alloc || hi.end_alloc < ptr) break; /* ubNTCYG[ */
			/* ptr.use͌Ă͂܂BgpubNA邱Ƃ͂肦܂B
			 * AgpubN̂́Ał͂ȂA4A炢̎ɍs悤łB
			 */
		}
		if(i != (int)hi.row) continue; /* ubN̓rŌs */

		return heapinfo_addr; /* ̃q[v͐ */
	}

	return 0; /* Ȃ */
}

/****************************************************************************
 *	heap_dump
 ****************************************************************************/

void
heap_dump(unsigned long heapinfo_addr)
{
	int i;
	unsigned long blktbl_addr;
	HEAPINFO hi;
	BLKTBL bt;

	/***** q[v̕\ *****/

	memcpy(&hi, &sram[heapinfo_addr - sysinfo.sram_top], sizeof(HEAPINFO));
	printf("============= q[v =============\n");
	printf("%08x: ansi_ucStartAlloc = %08x\n", heapinfo_addr +  0, hi.start_alloc);
	printf("%08x: ansi_ucEndAlloc   = %08x\n", heapinfo_addr +  4, hi.end_alloc  );
	printf("%08x: ansi_ucNxtAlcP    = %08x\n", heapinfo_addr +  8, hi.nxt_alc_p  );
	printf("%08x: ansi_ucTblPtr     = %08x\n", heapinfo_addr + 12, hi.tbl_ptr    );
	printf("%08x: ansi_ulRow        = %08x\n", heapinfo_addr + 16, hi.row        );

	/***** ubN̕\ *****/

	printf("============ ubN ============\n");
	printf("Index     Address   Size      Status  \n");
	for(i = 0; i < (int)hi.row; i++) {
		blktbl_addr = hi.end_alloc - sizeof(BLKTBL) * (i + 1);
		memcpy(&bt, &sram[blktbl_addr - sysinfo.sram_top], sizeof(BLKTBL));
		printf("%08x  %08x  %08x  %-8s\n", i, bt.ptr, bt.len, bt.use ? "USE" : "FREE");
	}
}
